-- Bunnymark - a cute benchmark
-- Copyright (C) 2023 Duane Robertson

-- main.lua

-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.


local BUNNY = 'art/bunny2.png'
local FONT_FILE = 'art/font/opensans-medium.ttf'
local FSIZE = 40
local GRASS = 'art/grass3.png'
local IHEIGHT = 1000
local IWIDTH = 720

local android
local b_img
local borders = {
	x = FSIZE,
	y = FSIZE,
	end_x = IWIDTH - FSIZE,
	end_y = IHEIGHT - FSIZE,
}
local bunnies = {}
local canv
local font
local fps = 0
local fpsl = {}
local label
local ripple
local sb
local scale = { x=1, y=1 }
local tim = 0
local tim2 = 0
local tr_off = { x=0, y=0 }
local trail


-- I'm still working on this. /sigh
local ripple_shader = [[
	uniform float TIME;

	const float tile = 0.5;
	const float aspect = 0.8;
	const vec2 time = vec2(1.0, 1.5);
	const vec2 offset = vec2(0.5, 0.5);
	const vec2 amplitude = vec2(0.003, 0.001);

	vec4 effect( vec4 color, Image tex, vec2 texture_coords, vec2 screen_coords )
    {
		vec2 tile_uv = texture_coords * tile;
		tile_uv.y *= aspect;

		vec2 wave;
		float scy = texture_coords.y * 20.0;
		wave.x += cos(TIME * time.x + scy + (tile_uv.y - tile_uv.x) * offset.x);
		wave.y += sin(TIME * time.x + scy + (tile_uv.x + tile_uv.y) * offset.y);

		return Texel(tex, tile_uv + wave * amplitude);
	}
]]


local function add_bunny(start_x, start_y)
	local b = {}
	b.image = b_img
	b.pos_x = start_x
	b.pos_y = start_y

	b.spd_x = math.random() - 0.5
	b.spd_y = math.random() - 0.5
	local n = math.sqrt(b.spd_x * b.spd_x + b.spd_y * b.spd_y)
	b.spd_x = (b.spd_x / n) * (math.random(300) + 100)
	b.spd_y = (b.spd_y / n) * (math.random(300) + 100)

	b.scale = math.random() * 2.0 + 0.5
	b.rotation = math.random() * 2 * 3.14
	b.ang_vel = (math.random() - 0.5) * 3.14
	b.color = {
		r = math.random() * 0.3 + 0.7,
		g = math.random() * 0.3 + 0.7,
		b = math.random() * 0.3 + 0.7,
	}

	bunnies[#bunnies+1] = b
end


function love.load()
	math.randomseed(os.time())

	local dx, dy = love.graphics.getDimensions()

	b_img = love.graphics.newImage(BUNNY)
	b_img:setFilter('nearest', 'nearest', 4)
	sb = love.graphics.newSpriteBatch(b_img)

	ripple = love.graphics.newShader(ripple_shader)

	canv = love.graphics.newCanvas(IWIDTH, IHEIGHT, { dpiscale = 1 })
	label = love.graphics.newCanvas(IWIDTH * 0.8, FSIZE * 4.2, { dpiscale = 1 })
	trail = love.graphics.newCanvas(IWIDTH * 2, IHEIGHT * 2, { dpiscale = 1 })

	local trail_img = love.graphics.newImage(GRASS)
	love.graphics.setCanvas(trail)
	for i = 0, trail:getWidth(), trail_img:getWidth() do
		for j = 0, trail:getHeight(), trail_img:getHeight() do
			love.graphics.draw(trail_img, i, j)
		end
	end
	love.graphics.setCanvas()

	font = love.graphics.newFont(FONT_FILE, FSIZE)

	android = (love.system.getOS() == 'Android')

	if android then
		local desk_x = math.min(love.window.getDesktopDimensions())
		local safe = dy / dx

		dx, dy = desk_x, math.floor(desk_x * safe)
		scale.x = dx / IWIDTH
		scale.y = dy / IHEIGHT
	end

	tr_off.x = trail:getWidth() / -4
	tr_off.y = trail:getHeight() / -4
end


local function print_text()
	love.graphics.setCanvas(label)
	love.graphics.clear()
	love.graphics.setColor(1, 1, 1, 0.5)
	love.graphics.rectangle('fill', 0, 0, IWIDTH * 0.8, FSIZE * 3.2, 20, 20)

	local fs = 0
	love.graphics.setFont(font)
	love.graphics.setColor(0, 0, 0)

	love.graphics.print('Tap anywhere to spawn', 20, fs)
	fs = fs + FSIZE
	love.graphics.print(#bunnies .. ' Total Bunnies', 20, fs)
	fs = fs + FSIZE
	love.graphics.print('Current FPS: ' .. tostring(fps), 20, fs)

	love.graphics.setCanvas()
end


function love.update(dt)
	tim = tim + dt
	tim2 = tim2 + dt
	if ripple:hasUniform('TIME') then
		ripple:send('TIME', tim2)
	end

	if tim > 1 then
		tim = 0
		fpsl[#fpsl+1] = love.timer.getFPS()
		while #fpsl > 10 do
			table.remove(fpsl, 1)
		end
		fps = 0
		for _, f in ipairs(fpsl) do
			fps = fps + f
		end
		fps = fps / #fpsl
		fps = math.floor(fps + 0.5)

		print_text()
	end

	for _, b in ipairs(bunnies) do
		b.pos_x = b.pos_x + b.spd_x * dt
		b.pos_y = b.pos_y + b.spd_y * dt
		b.rotation = math.fmod(b.rotation + b.ang_vel * dt, 2 * 3.14)

		local bounce

		if b.pos_x > borders.end_x then
			b.spd_x = math.abs(b.spd_x) * -1
			b.pos_x = borders.end_x
			bounce = true
		elseif b.pos_x < borders.x then
			b.spd_x = b.spd_x * -1
			b.pos_x = borders.x
			bounce = true
		end

		if b.pos_y > borders.end_y then
			b.spd_y = math.abs(b.spd_y) * -1
			b.pos_y = borders.end_y
			bounce = true
		elseif b.pos_y < borders.y then
			b.spd_y = b.spd_y * -1
			b.pos_y = borders.y
			bounce = true
		end

		if bounce and math.random(2) == 1 then
			b.spd_y = b.spd_y + (math.random() - 0.5) * 100
			b.spd_x = b.spd_x + (math.random() - 0.5) * 100
		end

		if b.spd_x > 500 then
			b.spd_x = 500
		end

		if b.spd_y > 500 then
			b.spd_y = 500
		end
	end
end


function love.draw()
	love.graphics.setCanvas(canv)
	love.graphics.setColor(1, 1, 1)
	love.graphics.clear()

	love.graphics.setShader(ripple)
	love.graphics.draw(trail, tr_off.x, tr_off.y)
	love.graphics.setShader()

	sb:clear()
	local ct = 0
	for _, b in ipairs(bunnies) do
		sb:setColor(b.color.r, b.color.g, b.color.b, 1.0)
		sb:add(b.pos_x, b.pos_y, b.rotation, b.scale, b.scale, 15, 20)
		ct = ct + 1

		if ct > 2000 then
			love.graphics.draw(sb)
			sb:clear()
			ct = 0
		end
	end

	if ct > 0 then
		love.graphics.draw(sb)
		sb:clear()
	end

	love.graphics.draw(label, FSIZE, FSIZE * 2)

	love.graphics.setCanvas()

	if android then
		love.graphics.draw(canv, 0, 0, 0, scale.x, scale.x)
	else
		love.graphics.draw(canv)
	end
end


function love.mousepressed(x, y, button)
    if button == 1 then
        for _ = 1, 500, 1 do
            add_bunny(x / scale.x, y / scale.x)
        end

		print_text()

	elseif button == 2 then
        for _ = 1, 500, 1 do
			table.remove(bunnies)
        end

		print_text()
    end
end



function love.run()
	if love.load then love.load(love.arg.parseGameArguments(arg), arg) end

	-- We don't want the first frame's dt to include time taken by love.load.
	if love.timer then love.timer.step() end

	local dt = 0

	-- Main loop time.
	return function()
		-- Process events.
		if love.event then
			love.event.pump()
			for name, a,b,c,d,e,f in love.event.poll() do
				if name == "quit" then
					if not love.quit or not love.quit() then
						return a or 0
					end
				end
				love.handlers[name](a,b,c,d,e,f)
			end
		end

		-- Update dt, as we'll be passing it to update
		if love.timer then dt = love.timer.step() end

		-- Call update and draw
		if love.update then love.update(dt) end -- will pass 0 if love.timer is disabled

		if love.graphics and love.graphics.isActive() then
			love.graphics.origin()
			love.graphics.clear(love.graphics.getBackgroundColor())

			if love.draw then love.draw() end

			love.graphics.present()
		end

		-- if love.timer then love.timer.sleep(0.001) end
	end
end
